#include "hal_dac63004.h"

#include "board_config.h"
#include "nrfx_spim.h"
#include "nrfx_ppi.h"
#include "nrfx_gpiote.h"
#include "error.h"
#include "log.h"

#define STEPS (30)

typedef struct {
    uint8_t data[STEPS][3];
} hal_dac63004_t;

static hal_dac63004_t hal_dac63004;
static const nrfx_spim_t spim3 = NRFX_SPIM_INSTANCE(0);

static void dac_data_write(uint8_t *d, const hal_dac63004_data_t *data)
{
    static const uint8_t channel_conv_table[] = { 0x19, 0x1a, 0x1b, 0x1c };
    d[0] = channel_conv_table[data->channel];
    d[1] = (data->value >> 2);
    d[2] = (data->value << 6);
}

static void spim_handler(nrfx_spim_evt_t const *p_event, void *p_context)
{
}

static uint32_t setup_spim(void)
{
    static const nrfx_spim_config_t config = {
        .sck_pin = BOARD_DAC63004_SCK_PIN,
        .mosi_pin = BOARD_DAC63004_MOSI_PIN,
        .miso_pin = NRFX_SPIM_PIN_NOT_USED,
        .ss_pin = BOARD_DAC63004_SS_PIN,
        .ss_active_high = false,
        .irq_priority = 6,
        .orc = 0xca,
        .frequency = NRF_SPIM_FREQ_8M,
        .mode = NRF_SPIM_MODE_0,
        .bit_order = NRF_SPIM_BIT_ORDER_MSB_FIRST,
    };
    nrfx_err_t err_code = nrfx_spim_init(&spim3, &config, spim_handler, NULL);
    assert_param(NRFX_SUCCESS == err_code);
    static const nrfx_spim_xfer_desc_t xfer = {
        .p_tx_buffer = (uint8_t const *) hal_dac63004.data,
        .tx_length = 3,
        .p_rx_buffer = NULL,
        .rx_length = 0,
    };
    err_code = nrfx_spim_xfer(&spim3, &xfer,
                              (NRFX_SPIM_FLAG_TX_POSTINC | NRFX_SPIM_FLAG_HOLD_XFER | NRFX_SPIM_FLAG_REPEATED_XFER |
                               NRFX_SPIM_FLAG_NO_XFER_EVT_HANDLER));
    return nrfx_spim_start_task_get(&spim3);
}

static uint32_t setup_trigger(void)
{
    static const nrfx_gpiote_in_config_t config = {
        .sense = NRF_GPIOTE_POLARITY_HITOLO,
        .pull = NRF_GPIO_PIN_PULLUP,
        .is_watcher = false,
        .hi_accuracy = true,
        .skip_gpio_setup = true,
    };
    nrfx_err_t err = nrfx_gpiote_init();
    nrf_gpio_cfg_input(BOARD_TRIGGER_IN_PIN, config.pull);
    assert_param(NRFX_SUCCESS == err);
    err = nrfx_gpiote_in_init(BOARD_TRIGGER_IN_PIN, &config, NULL);
    assert_param(NRFX_SUCCESS == err);
    nrfx_gpiote_in_event_enable(BOARD_TRIGGER_IN_PIN, false);
    return nrfx_gpiote_in_event_addr_get(BOARD_TRIGGER_IN_PIN);
}

static void setup_ppi(uint32_t tep, uint32_t eep)
{
    nrf_ppi_channel_t channel;
    nrfx_err_t err = nrfx_ppi_channel_alloc(&channel);
    assert_param(NRFX_SUCCESS == err);
    err = nrfx_ppi_channel_assign(channel, eep, tep);
    assert_param(NRFX_SUCCESS == err);
    err = nrfx_ppi_channel_enable(channel);
    assert_param(NRFX_SUCCESS == err);
}

void hal_dac63004_init(void)
{
    const uint32_t tep = setup_spim();
    const uint32_t eep = setup_trigger();
    setup_ppi(tep, eep);
}

void hal_dac63004_send(const hal_dac63004_data_t *data, size_t len)
{
    assert_param(data != NULL);
    assert_param(len <= STEPS);
    for (size_t k = 0; k < len; k++) {
        dac_data_write(hal_dac63004.data[k], &data[k]);
    }
    spim3.p_reg->TXD.PTR = (uintptr_t) hal_dac63004.data[0];
}
